# MINERVA AI-Lab
# Institute of Computer Engineering
# University of Seville, Spain
#
# Copyright 2026 Salvador de la Torre Gonzalez
# Antonio Bello Castro,
# José M. Núñez Portero
#
# Developed and currently maintained by:
#    Salvador de la Torre Gonzalez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#     SPDX-License-Identifier: Apache-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
GAUFS: Genetic Algorithm for Unsupervised Feature Selection for Clustering.

This package provides the main GAUFS interface and core public classes.

The API exposes:
- GAUFS: the main class for running the algorithm
- DataGenerator: two synthetic data generators for clustering experiments
- `read_unlabeled_data_csv` and `read_labeled_data_csv`: functions to load datasets from CSV files

Additional functionality, such as clustering experiments and evaluation
metrics, is available through dedicated subpackages.
"""

from .gaufs import Gaufs
from .data_generator import DataGenerator
from .utils import read_unlabeled_data_csv, read_labeled_data_csv

__all__ = ["Gaufs", "DataGenerator", "read_unlabeled_data_csv", "read_labeled_data_csv"]



# MINERVA AI-Lab
# Institute of Computer Engineering
# University of Seville, Spain
#
# Copyright 2026 Salvador de la Torre Gonzalez
# Antonio Bello Castro,
# José M. Núñez Portero
#
# Developed and currently maintained by:
#    Salvador de la Torre Gonzalez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#     SPDX-License-Identifier: Apache-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Internal utility functions for evaluation metrics.

This module contains helper functions used internally by the library.
"""

import pandas as pd


# Contingency table
@staticmethod
def _get_cont_table(c1, c2, round=True):
    """
    Generate a contingency table (cross-tabulation) between two categorical variables.
    """
    res = pd.crosstab(c1, c2, normalize="index") * 100
    return res.round(1) if round else res



# MINERVA AI-Lab
# Institute of Computer Engineering
# University of Seville, Spain
#
# Copyright 2026 Salvador de la Torre Gonzalez
# Antonio Bello Castro,
# José M. Núñez Portero
#
# Developed and currently maintained by:
#    Salvador de la Torre Gonzalez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#     SPDX-License-Identifier: Apache-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Genetic Search implementation for GAUFS.

This module contains the internal Genetic Algorithm implementation used
by GAUFS to optimize feature selection.

.. note::
   This class and its methods are for internal use only and are
   not exposed as part of the public API.
"""

import os
import random
from concurrent.futures import ProcessPoolExecutor

import numpy as np
import matplotlib.pyplot as plt
from deap import base, tools, algorithms, creator

from .utils import _evaluate_ind


class GeneticSearch:
    def __init__(
        self,
        unlabeled_data,
        ngen,
        npop,
        cxpb,
        cxpb_rest_of_genes,
        mutpb,
        convergence_generations,
        hof_size,
        clustering_method,
        evaluation_metric,
        cluster_number_search_band,
        verbose,
        path_store_log=None,
        path_store_plot=None,
        seed=random.randint(0, 10000),
    ):
        self.unlabeled_data = unlabeled_data.copy()
        self.num_var = unlabeled_data.shape[1]
        self.ngen = ngen
        self.npop = npop
        self.cxpb = cxpb
        self.cxpb_rest_of_genes = cxpb_rest_of_genes
        self.mutpb = mutpb
        self.convergence_generations = convergence_generations
        self.hof_size = hof_size
        self.clustering_method = clustering_method
        self.evaluation_metric = evaluation_metric
        self.cluster_number_search_band = cluster_number_search_band
        self.verbose = verbose
        # If None then no log will be stored
        self.path_store_log = path_store_log
        self.store_log = path_store_log is not None
        # If None no graph will be plotted
        self.graph_evolution = path_store_plot is not None
        self.path_store_plot = path_store_plot
        self.seed = seed

        # Hall of Fame with the best chromosomes found (num_clusters + variable_selection)
        self.hof = None

        # dictionary={num_cluster:(max_fitness asociated,chromosome_where_it_was_achieved)}
        self.num_clusters_and_its_max_fitness = dict()
        # dictionary={variable_selection:(maximum fitness found for that selection (regardeless of the cluster number), number of times a chromosome with this selection entered in the hof)}
        self.hof_counter = dict()
        self.hof_weighted = dict()

        # logbook to store statistics
        self.logbook = tools.Logbook()
        self.logbook.header = "gen", "nevals", "avg", "std", "min", "max"

    ## Function for encoding individuals in the genetic algorithm
    def _init_individual(self, container, attr_bool, data):
        """
        For initializing individuals in the genetic algorithm.
        First gene: number of clusters (random number within cluster_number_search_band (min_inclusive, max_exclusive))
        Remaining genes: binary (each gene i represents whether the i-th variable is considered for clustering)
        """
        ## To initialize the number of clusters within the desired range
        num_clusters_range = range(
            self.cluster_number_search_band[0], self.cluster_number_search_band[1]
        )
        k = random.choice(num_clusters_range)

        # Creation of the rest of the individual with binary genes
        individual = container([k] + [attr_bool() for _ in range(self.num_var)])
        return individual

    ### Adapted definition of genetic operators
    def _cxUniformModified(self, ind1, ind2):
        """
        Uniform crossover with exchange of the first gene (number of clusters)
        and the rest of the binary genes.
        """
        # Crossover of the first gene (number of clusters)
        ind1[0], ind2[0] = ind2[0], ind1[0]

        # Uniform crossover for the rest of the genes
        for i in range(1, len(ind1)):
            # Exchange occurs in the remaining genes with probability cxpb_rest_of_genes
            if random.random() < self.cxpb_rest_of_genes:
                ind1[i], ind2[i] = ind2[i], ind1[i]
        return ind1, ind2

    def _mutFlipBitModified(self, ind):
        """
        Modified bit-by-bit mutation so that the number of clusters
        does not go outside the established range
        """
        mutation_type = random.choice([1, 2, 3])
        if mutation_type == 1:
            num_clusters_range = range(
                self.cluster_number_search_band[0], self.cluster_number_search_band[1]
            )
            # We choose a new number of clusters within the allowed range
            ind[0] = random.choice(num_clusters_range)
        elif mutation_type == 2:
            i = random.choice(range(1, len(ind)))
            # Usual bit-flip mutation
            ind[i] = 1 - ind[i]
        else:
            num_clusters_range = range(
                self.cluster_number_search_band[0], self.cluster_number_search_band[1]
            )
            # We choose a new number of clusters within the allowed range
            ind[0] = random.choice(num_clusters_range)
            i = random.choice(range(1, len(ind)))
            # Usual bit-flip mutation
            ind[i] = 1 - ind[i]
        # A tuple must be returned
        return (ind,)

    def _evaluate_individual(self, individual):
        num_clusters = individual[0]
        variables = individual[1:]
        # We need to return a tuple
        return (
            _evaluate_ind(
                self.unlabeled_data,
                num_clusters,
                variables,
                self.clustering_method,
                self.evaluation_metric,
            ),
        )

    def _map_function(self, func, *args):
        # create as many workers as cores available.
        # The overhead of creating a pool each time is very low.
        with ProcessPoolExecutor() as executor:
            return list(executor.map(func, *args))

    def run(self):
        # Check that the number of clusters search band is valid
        if self.cluster_number_search_band[0] >= self.cluster_number_search_band[1]:
            raise ValueError(
                "The cluster_number_search_band is not valid. The first element must be less than the second one."
            )
        if self.cluster_number_search_band[1] > self.unlabeled_data.shape[0]:
            raise ValueError(
                "The cluster_number_search_band is not valid. The maximum number of clusters cannot be greater than the number of data points."
            )

        random.seed(self.seed)

        # Delete existing classes if they exist to avoid RuntimeWarning
        if hasattr(creator, "Fitness"):
            del creator.Fitness
        if hasattr(creator, "Individual"):
            del creator.Individual

        # Definition of the optimization problem type
        creator.create("Fitness", base.Fitness, weights=(1.0,))
        creator.create("Individual", list, fitness=creator.Fitness)

        # Initialization of Deap's toolbox
        toolbox = base.Toolbox()

        # Creation of each boolean gene of a chromosome
        toolbox.register("attr_bool", random.randint, 0, 1)

        # Definition of each chromosome (individual)
        toolbox.register(
            "individual",
            lambda: self._init_individual(
                creator.Individual, toolbox.attr_bool, self.unlabeled_data
            ),
        )

        # The population is a list of individuals
        toolbox.register("population", tools.initRepeat, list, toolbox.individual)

        # Register the evaluation function, crossover, mutation, and selection operators
        toolbox.register("evaluate", self._evaluate_individual)
        toolbox.register("mate", self._cxUniformModified)
        toolbox.register("mutate", self._mutFlipBitModified)
        toolbox.register("select", tools.selTournament, tournsize=10)

        # Parallel evaluation to speed up computations
        toolbox.register("map", self._map_function)

        # Fitness statistics per generation
        stats_fit = tools.Statistics(lambda ind: ind.fitness.values)

        # The mean, standard deviation, minimum, and maximum fitness per generation will be shown
        stats_fit.register("avg", np.mean, axis=0)
        stats_fit.register("std", np.std, axis=0)
        stats_fit.register("min", np.min, axis=0)
        stats_fit.register("max", np.max, axis=0)

        # Creation of an initial population
        population = toolbox.population(self.npop)

        # List that will store the best individuals generation by generation
        best_inds = []
        # List that will store the average fitness of individuals in each generation
        avg_fitness_history = []

        # Initialization of the HallOfFame
        self.hof = tools.HallOfFame(self.hof_size)

        # Initial evaluation of individuals
        invalid_ind = [ind for ind in population if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)

        # update the num_clusters_and_its_max_fitness dictionary with the maximum fitness for each cluster number
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit
            val = fit[0]
            cluster_number = ind[0]
            if cluster_number in self.num_clusters_and_its_max_fitness:
                if self.num_clusters_and_its_max_fitness[cluster_number][0] < val:
                    self.num_clusters_and_its_max_fitness[cluster_number] = (val, ind)
            else:
                self.num_clusters_and_its_max_fitness[cluster_number] = (val, ind)

        # Update the Hall of Fame with the best individuals
        self.hof.update(population)

        # Update the hof_counter dictionary with the best variable selections found in the hof regardeless of the cluster number
        for ind in self.hof:
            variables_selection = tuple(ind[1:])
            if variables_selection in self.hof_counter:
                old = self.hof_counter[variables_selection]
                # we store the maximum fitness found for that variable selection and the number of times it has been present in the hof
                self.hof_counter[variables_selection] = (
                    max(old[0], ind.fitness.values[0]),
                    old[1] + 1,
                )
            else:
                self.hof_counter[variables_selection] = (ind.fitness.values[0], 1)

        # Definition of variables to search for convergence in the HoF
        hof_unchanged_count = 0
        latest_hof_snapshot = set(tuple(ind) for ind in self.hof)

        # Update the lists for best individuals and average fitness per generation
        if self.graph_evolution:
            best_inds.append(self.hof[0])  # Inicialmente
            avg_fitness_history.append(
                np.mean([ind.fitness.values[0] for ind in population])
            )

        # Register statistics of the initial population and print them
        record = stats_fit.compile(population)
        self.logbook.record(gen=0, nevals=len(invalid_ind), **record)
        if self.verbose:
            print(self.logbook.stream)

        # Run the evolutionary algorithm
        for gen in range(self.ngen):

            # Select the next generation of individuals
            offspring = toolbox.select(population, len(population) - self.hof_size)

            # Mutation and reproduction of individuals
            offspring = algorithms.varAnd(offspring, toolbox, self.cxpb, self.mutpb)

            # Evaluation of the individuals with invalid fitness which are the new ones
            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
            fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)

            # update the num_clusters_and_its_max_fitness dictionary with the maximum fitness for each cluster number
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit
                val = fit[0]
                cluster_number = ind[0]
                if cluster_number in self.num_clusters_and_its_max_fitness:
                    if self.num_clusters_and_its_max_fitness[cluster_number][0] < val:
                        self.num_clusters_and_its_max_fitness[cluster_number] = (
                            val,
                            ind,
                        )
                else:
                    self.num_clusters_and_its_max_fitness[cluster_number] = (val, ind)

            # Elitism: Extend the population with the individuals from the Hall of Fame
            offspring.extend(self.hof.items)

            # Update the Hall of Fame with the best individuals found so far
            self.hof.update(offspring)

            # Update the hof_counter dictionary with the best variable selections found in the hof regardeless of the cluster number
            for ind in self.hof:
                variables_selection = tuple(ind[1:])
                if variables_selection in self.hof_counter:
                    old = self.hof_counter[variables_selection]
                    self.hof_counter[variables_selection] = (
                        max(old[0], ind.fitness.values[0]),
                        old[1] + 1,
                    )  # maximo fitness y cuento las veces que ha entrado
                else:
                    self.hof_counter[variables_selection] = (ind.fitness.values[0], 1)

            # If we are in the last generation, store the individuals and their fitness values
            if gen == self.ngen - 1:
                self.last_gen = {tuple(ind): ind.fitness.values[0] for ind in offspring}

            # Replace the current population with the offspring
            population[:] = offspring

            # Update the lists for best individuals and average fitness per generation
            if self.graph_evolution:
                best_inds.append(self.hof[0])
                avg_fitness_history.append(
                    np.mean([ind.fitness.values[0] for ind in population])
                )

            record = stats_fit.compile(population)
            self.logbook.record(gen=gen + 1, nevals=len(invalid_ind), **record)

            # Output and store of statistics
            if self.verbose:
                print(self.logbook.stream)
            if self.store_log:
                self.save_log_to_file(early_stopped=False, current_gen=self.ngen - 1)

            # if the hof has not changed in the last convergence_generations generations, we stop the algorithm
            current_hof_snapshot = set(tuple(ind) for ind in self.hof)
            if current_hof_snapshot == latest_hof_snapshot:
                hof_unchanged_count += 1
            else:
                hof_unchanged_count = 0
                latest_hof_snapshot = current_hof_snapshot
            if hof_unchanged_count >= self.convergence_generations:
                if self.verbose:
                    print(
                        f"Early Stopping due to Hall Of Fame not changing in {gen+1} generations"
                    )
                self.last_gen = {tuple(ind): ind.fitness.values[0] for ind in offspring}
                if self.store_log:
                    self.save_log_to_file(early_stopped=True, current_gen=gen)
                break

        # Plot the evolution of the best and average fitness score
        if self.graph_evolution:
            self.save_graph_evolution(best_inds, avg_fitness_history)
        if self.store_log:
            # Save Hall of Fame from the last genetic execution to a file
            with open(os.path.join(self.path_store_log, "hall_of_fame.txt"), "w") as f:
                for ind in self.hof:
                    f.write(f"Individual: {ind}, Fitness: {ind.fitness.values[0]}\n")
            # Save hof_counter to a file
            with open(
                os.path.join(self.path_store_log, "hall_of_fame_counter.txt"), "w"
            ) as f:
                for key, value in self.hof_counter.items():
                    f.write(
                        f"Selection: {key}, (Max Fitness: {value[0]}, Times in HoF: {value[1]})\n"
                    )

        return self.hof_counter, self.num_clusters_and_its_max_fitness

    def save_graph_evolution(self, best_inds, avg_fitness_history):
        fitness_vals = []
        for _, ind in enumerate(best_inds, start=1):
            fitness_vals.append(ind.fitness.values[0])
        num_generations = len(fitness_vals)
        plt.figure(figsize=(10, 5))
        plt.plot(
            range(1, num_generations + 1),
            fitness_vals,
            marker="o",
            linestyle="-",
            color="b",
            label="Fitness of the best individual",
        )
        plt.plot(
            range(1, num_generations + 1),
            avg_fitness_history[:num_generations],
            marker="x",
            linestyle="--",
            color="r",
            label="Average fitness of the population",
        )
        plt.title("Fitness Evolution Over Generations")
        plt.xlabel("Number of Generations")
        plt.ylabel("Fitness")
        plt.legend()
        plt.grid(True)
        plt.savefig(
            os.path.join(self.path_store_plot, "fitness_evolution.png"),
            dpi=300,
            bbox_inches="tight",
        )
        plt.close()

    def save_log_to_file(self, current_gen, early_stopped):
        """Save the logbook to a file."""
        # Create directory if it doesn't exist
        os.makedirs(self.path_store_log, exist_ok=True)

        with open(
            os.path.join(self.path_store_log, "genetic_algorithm_log.txt"), "w"
        ) as f:
            if early_stopped:
                f.write(f"Early stopping at generation {current_gen + 1}\n\n")
            f.write(str(self.logbook))



# MINERVA AI-Lab
# Institute of Computer Engineering
# University of Seville, Spain
#
# Copyright 2026 Salvador de la Torre Gonzalez
# Antonio Bello Castro,
# José M. Núñez Portero
#
# Developed and currently maintained by:
#    Salvador de la Torre Gonzalez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#     SPDX-License-Identifier: Apache-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
hierarchical.py

Hierarchical clustering experiment for GAUFS.
Implements the HierarchicalExperiment class, which
extends ClusteringExperiment using AgglomerativeClustering.
"""

from sklearn.cluster import AgglomerativeClustering

from .base import ClusteringExperiment


class HierarchicalExperiment(ClusteringExperiment):
    """
    Hierarchical clustering experiment using Agglomerative Clustering.

    Parameters
    ----------
    unlabeled_data : pandas.DataFrame or pandas.Series, optional
        The unlabeled dataset to be clustered. If provided, a copy is stored.
    linkage : str, default='ward'
        Linkage criterion for hierarchical clustering. Options: 'ward',
        'complete', 'average', 'single'.

    Attributes
    ----------
    n_clusters : int or None
        Number of clusters to form. Must be set before calling run().
    linkage : str
        The linkage method used for clustering.
    """

    def __init__(self, unlabeled_data=None, linkage="ward"):
        """
        Initialize a HierarchicalExperiment instance.

        Parameters
        ----------
        unlabeled_data : pandas.DataFrame or pandas.Series, optional
            The unlabeled dataset to be clustered.
        linkage : str, default='ward'
            Linkage criterion for hierarchical clustering.

        """
        if unlabeled_data is not None:
            self.unlabeled_data = unlabeled_data.copy()
        self.n_clusters = None
        # Linkage method: 'ward', 'complete', 'average', 'single'
        self.linkage = linkage

    def run(self):
        """
        Execute hierarchical clustering with AgglomerativeClustering.

        Creates an AgglomerativeClustering instance with the specified
        number of clusters and linkage method, then runs the clustering.

        """
        self.algorithm = AgglomerativeClustering(
            n_clusters=self.n_clusters, linkage=self.linkage
        )
        super().run()



# MINERVA AI-Lab
# Institute of Computer Engineering
# University of Seville, Spain
#
# Copyright 2026 Salvador de la Torre Gonzalez
# Antonio Bello Castro,
# José M. Núñez Portero
#
# Developed and currently maintained by:
#    Salvador de la Torre Gonzalez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#     SPDX-License-Identifier: Apache-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Evaluation metrics for clustering algorithms in GAUFS.

This subpackage provides internal and external clustering evaluation metrics,
as well as the EvaluationMetric base class for implementing custom metrics.
"""

from .base import EvaluationMetric

from .internal import (
    SilhouetteScore,
    CalinskiHarabaszScore,
    DaviesBouldinScore,
    DaviesBouldinScoreForMaximization,
    DunnScore,
    SSEScore,
    SSEScoreForMaximization,
)

from .external import (
    AdjustedMutualInformationScore,
    AdjustedRandIndexScore,
    VMeasureScore,
    FowlkesMallowsScore,
    FScore,
    NMIScore,
    HScore,
    Chi2,
    DobPertScore,
)

__all__ = [
    "EvaluationMetric",
    "SilhouetteScore",
    "CalinskiHarabaszScore",
    "DaviesBouldinScore",
    "DaviesBouldinScoreForMaximization",
    "DunnScore",
    "SSEScore",
    "SSEScoreForMaximization",
    "AdjustedMutualInformationScore",
    "AdjustedRandIndexScore",
    "VMeasureScore",
    "FowlkesMallowsScore",
    "FScore",
    "NMIScore",
    "HScore",
    "Chi2",
    "DobPertScore",
]



# MINERVA AI-Lab
# Institute of Computer Engineering
# University of Seville, Spain
#
# Copyright 2026 Salvador de la Torre Gonzalez
# Antonio Bello Castro,
# José M. Núñez Portero
#
# Developed and currently maintained by:
#    Salvador de la Torre Gonzalez
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#     SPDX-License-Identifier: Apache-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Internal utilities for GAUFS.

This module contains helper functions used internally by the library.
These functions are not part of the public API, except for data loading.

In addition, it provides the following functions for users:
- `read_unlabeled_data_csv`
- `read_labeled_data_csv`
"""

import pandas as pd
import numpy as np
import copy
import os
import shutil


# ----------------------------
# Public functions for users
# ----------------------------
def read_unlabeled_data_csv(filepath):
    """
    Reads unlabeled data from a CSV file.

    Accepts CSVs with or without a header.

    Parameters
    ----------
    filepath : str
        Path to the CSV file.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the unlabeled data of shape (n_samples, n_features).

    """
    # Try reading with header
    df = pd.read_csv(filepath)

    # If the first row is numeric, there was no real header
    if df.columns.to_list()[0].startswith("Unnamed") or all(
        c.replace(".", "", 1).isdigit() for c in df.columns
    ):
        df = pd.read_csv(filepath, header=None)

    # Force numeric (important for ML pipelines)
    df = df.apply(pd.to_numeric, errors="raise")
    return df


def read_labeled_data_csv(filepath):
    """
    Reads labeled data from a CSV file.

    Accepts CSVs with or without a header. The format of the labeled data CSV is:
    - First n-1 columns: features
    - Last column: true labels

    Parameters
    ----------
    filepath : str
        Path to the CSV file.

    Returns
    -------
    unlabeled_data : pd.DataFrame
        DataFrame of shape (n_samples, n_features) containing the features.
    true_labels : np.ndarray
        Array of shape (n_samples,) containing the true labels.

    """
    df = pd.read_csv(filepath)

    # If the first row is numeric, there was no real header
    if df.columns.to_list()[0].startswith("Unnamed") or all(
        c.replace(".", "", 1).isdigit() for c in df.columns
    ):
        df = pd.read_csv(filepath, header=None)

    # Force numeric (important for ML pipelines)
    df = df.apply(pd.to_numeric, errors="raise")

    unlabeled_data = df.iloc[:, :-1]
    true_labels = df.iloc[:, -1].values

    return unlabeled_data, true_labels


# ----------------------------
# Internal helper functions (private)
# ----------------------------
def _evaluate_ind(
    unlabeled_data, cluster_number, variables, clustering_method, evaluation_metric
):
    """
    Evaluate an individual solution in the GAUFS algorithm.

    Evaluates a specific combination of cluster number and feature selection
    by performing clustering and computing the evaluation metric.

    Parameters
    ----------
    unlabeled_data : pd.DataFrame
        The unlabeled data to cluster.
    cluster_number : int
        The number of clusters to form.
    variables : list
        A binary list indicating selected variables (1 for selected, 0 for not).
    clustering_method : ClusteringExperiment
        An instance of a clustering method with a ``run()`` method.
    evaluation_metric : EvaluationMetric
        An instance of an evaluation metric with a ``compute()`` method.

    Returns
    -------
    float
        The evaluation score of the clustering result.
    """
    try:
        # if no variables are selected, return a very low fitness
        if np.all(np.array(variables) == 0):
            return -10000000000

        filtered_vars = [
            var for var, i in zip(unlabeled_data.columns, variables) if i == 1
        ]
        filtered_data = unlabeled_data[filtered_vars]

        # Create a copy of the provided clustering experiment
        # deepcopying should not be necessary because clustering_method.unlabeled_data
        # should be None if running the GA, but to be safe we do it in case this method is used elsewhere
        experiment = copy.deepcopy(clustering_method)
        # copy the filtered unlabeled data
        experiment.set_unlabeled_data(filtered_data)
        # Set the number of clusters
        experiment.n_clusters = cluster_number
        # Run the clustering algorithm
        experiment.run()
        # get the results
        assigned_clusters = experiment.assigned_clusters

        # Evaluate the clustering result with the provided metric and the filtered data
        ev = evaluation_metric.compute(
            assigned_clusters=assigned_clusters, unlabeled_data=filtered_data
        )

        return ev

    except Exception as e:
        print(
            f"Error evaluating the individual with {cluster_number} clusters and the selection {variables}; Exception: {type(e).__name__} - {e}"
        )


def _compute_variable_significance(
    num_variables, hof_counter, max_number_selections_for_ponderation
):
    """
    Compute variable significance from Hall of Fame individuals.

    Calculates variable significance values between 0 and 1 based on the
    evaluation scores and selection frequency of individuals in the Hall of Fame.

    Parameters
    ----------
    num_variables : int
        Number of variables in the dataset.
    hof_counter : dict
        Dictionary mapping variable selections to their statistics.
        Keys are binary tuples representing variable selections.
        Values are tuples of (score, selection_count) where:

        - score : Best fitness achieved for a chromosome with that variable selection
        - selection_count : Number of times a chromosome with that selection entered the Hall of Fame

    max_number_selections_for_ponderation : int
        Maximum number of top individuals to consider for weight computation.

    Returns
    -------
    np.ndarray
        Variable significance values between 0 and 1 for each variable.
    """
    selections = sorted(hof_counter.items(), key=lambda item: item[1][0], reverse=True)[
        :max_number_selections_for_ponderation
    ]
    scores = []
    total = 0
    for it in selections:
        # the score of the selection if the maximun fitness achieved by that selection multiplied by the number of times it was selected in the HoF
        score = it[1][0] * it[1][1]
        scores.append(score)
        total += score

    # normalize scores so that they sum to 1
    scores_normalized = [x / total for x in scores]
    res = [0] * num_variables
    for i in range(num_variables):
        for j, s in enumerate(scores_normalized):
            res[i] += s * selections[j][0][i]
    return res


def _clear_directory(directory_path):
    """
    Clears all contents of the specified directory.
    Works on Windows and Linux.
    """
    if os.path.exists(directory_path):
        try:
            shutil.rmtree(directory_path)
        except PermissionError:
            raise PermissionError(
                f"Could not delete {directory_path}. Check file permissions or open handles."
            )

    os.makedirs(directory_path, exist_ok=True)


# Helper function to plot with discontinuities
def _plot_discontinuous(ax, x_vals, y_vals, **kwargs):
    """Plot lines only between consecutive x values"""
    i = 0
    while i < len(x_vals):
        # Find the end of the consecutive sequence
        j = i
        while j < len(x_vals) - 1 and x_vals[j + 1] == x_vals[j] + 1:
            j += 1

        # Plot this consecutive segment
        ax.plot(x_vals[i : j + 1], y_vals[i : j + 1], **kwargs)

        # Move to next segment
        i = j + 1


def _convert_to_serializable(obj):
    """
    Function to convert numpy arrays to lists
    """
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    elif isinstance(obj, dict):
        return {key: _convert_to_serializable(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [_convert_to_serializable(item) for item in obj]
    elif isinstance(obj, (np.integer, np.floating)):
        return obj.item()
    return obj


def _get_variables_over_threshold(variables_weights, threshold):
    """
    Given a list of variable weights and a threshold, returns a binary list indicating which variables have weights greater than or equal to the threshold.
    """
    return [1 if w >= threshold else 0 for w in variables_weights]


def _get_dictionary_num_clusters_fitness(
    unlabeled_data,
    variable_selection,
    cluster_number_search_band,
    clustering_method,
    evaluation_metric,
):
    """
    Computes a dictionary mapping the number of clusters whithin the cluster_number_search_band (min_inclusive, max_exclusive) to their corresponding fitness scores given the fixed variable selection binary_variable_selection.
    """
    dict_clusters_fitness = {}
    for k in range(cluster_number_search_band[0], cluster_number_search_band[1]):
        dict_clusters_fitness[k] = _evaluate_ind(
            unlabeled_data=unlabeled_data,
            cluster_number=k,
            variables=variable_selection,
            clustering_method=clustering_method,
            evaluation_metric=evaluation_metric,
        )
    return dict_clusters_fitness


# The overhead of not computing the maximum in _get_dictionary_num_clusters_fitness is negligible compared to the clustering computations
def _get_num_clusters_with_best_fitness(dict_clusters_fit):
    """
    Given a dictionary mapping number of clusters to fitness scores, returns the number of clusters with the best fitness.
    In case of tie, returns the highest number of clusters among those with the best fitness.
    """
    num_clusters_for_maximum_fitness = None
    max_fitness = None
    for key in dict_clusters_fit.keys():
        # In case of tie we select the highest number of clusters
        if (
            max_fitness == None
            or max_fitness < dict_clusters_fit[key]
            or (
                max_fitness == dict_clusters_fit[key]
                and num_clusters_for_maximum_fitness < key
            )
        ):
            max_fitness = dict_clusters_fit[key]
            num_clusters_for_maximum_fitness = key
    return num_clusters_for_maximum_fitness, max_fitness


def _min_max_normalize_dictionary(dictionary):
    """
    Returns a new dictionary with the values min-max normalized to the range [0, 1].
    """
    values = list(dictionary.values())
    max_value = max(values)
    min_value = min(values)
    # In case all values are the same, return a dictionary with all values set to 0.0
    if max_value == min_value:
        return {k: 0.0 for k in dictionary.keys()}

    return {k: (v - min_value) / (max_value - min_value) for k, v in dictionary.items()}
